package hu.kozadat.search.base.preprocessor;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * User: codex
 */
@Component("centuryPreprocessor")
public class CenturyPreprocessor extends QueryPreprocessor {

    private static final Pattern centuryPattern = Pattern.compile("(^|[\\s()])((isz|ie)[.]?[\\s]*)?([0-9]{1,2}|[ivx]{1,4})[.]?[\\s]*(század|sz[.]?)([\\s]+(eleje|közepe|vége|első[\\s]fele|második[\\s]fele))?(\\b|\\B)", Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE);

    private static final Map<String, String> centuryConversionMap;

    public static final String CENTURY_BEGINNING = "eleje";
    public static final String CENTURY_MIDDLE = "közepe";
    public static final String CENTURY_END = "vége";
    public static final String CENTURY_FIRST_HALF = "első fele";
    public static final String CENTURY_SECOND_HALF = "második fele";

    static {
        centuryConversionMap = new HashMap<String, String>();
        centuryConversionMap.put("i", "1");
        centuryConversionMap.put("ii", "2");
        centuryConversionMap.put("iii", "3");
        centuryConversionMap.put("iv", "4");
        centuryConversionMap.put("v", "5");
        centuryConversionMap.put("vi", "6");
        centuryConversionMap.put("vii", "7");
        centuryConversionMap.put("viii", "8");
        centuryConversionMap.put("ix", "9");
        centuryConversionMap.put("x", "10");
        centuryConversionMap.put("xi", "11");
        centuryConversionMap.put("xii", "12");
        centuryConversionMap.put("xiii", "13");
        centuryConversionMap.put("xiv", "14");
        centuryConversionMap.put("xv", "15");
        centuryConversionMap.put("xvi", "16");
        centuryConversionMap.put("xvii", "17");
        centuryConversionMap.put("xviii", "18");
        centuryConversionMap.put("xix", "19");
        centuryConversionMap.put("xx", "20");
        centuryConversionMap.put("xxi", "21");
    }

    private DatePrecision datePrecision;

    public DatePrecision getDatePrecision() {
        return datePrecision;
    }

    @Autowired
    public void setDatePrecision(DatePrecision datePrecision) {
        this.datePrecision = datePrecision;
    }

    private String formatStartYear(int year) {
        GregorianCalendar cal = new GregorianCalendar(year, 0, 1);
        Date event = cal.getTime();

        if (datePrecision != null)
            event = datePrecision.adjustDateDownwards(event);

        return df.format(event);
    }

    private String formatEndYear(int year) {
        GregorianCalendar cal = new GregorianCalendar(year, 11, 31);
        Date event = cal.getTime();

        if (datePrecision != null)
            event = datePrecision.adjustDateUpwards(event);

        return df.format(event);
    }

    public QueryPreprocessorResult preprocess(String segment, boolean appendFieldName) {
        QueryPreprocessorResult res = new QueryPreprocessorResult();
        res.setOriginalQuery(segment);
        Matcher centuryRangeMatcher = centuryPattern.matcher(segment);
        StringBuffer sb = new StringBuffer();

        String fieldStr = appendFieldName ? "date:" : "";

        while (centuryRangeMatcher.find()) {
            //--  not null
            String startStr = centuryRangeMatcher.group(1);
            //-- can be null
            String timelinePartSelector = com.actiwise.einfoszab.Utils.null2String(centuryRangeMatcher.group(3)).toLowerCase();
            //-- not null
            String centuryStr = com.actiwise.einfoszab.Utils.null2String(centuryRangeMatcher.group(4)).toLowerCase();
            //-- can be null
            String centuryPartStr = com.actiwise.einfoszab.Utils.null2String(centuryRangeMatcher.group(7)).toLowerCase();

            boolean isIsz = true;
            if (!"".equals(timelinePartSelector)) {
                if (BEFORE_CHRISTMAS.equals(timelinePartSelector))
                    isIsz = false;
            }

            if (centuryConversionMap.get(centuryStr) != null)
                centuryStr = centuryConversionMap.get(centuryStr);

            int century = 0;
            boolean validCentury = true;
            try {
                century = Integer.parseInt(centuryStr);

            } catch (NumberFormatException nex) {
                validCentury = false;
            }

            String modifiedCentury;

            if (validCentury) {
                int startYear = (century - 1) * 100 + 1;
                int endYear = century * 100;
                ;

                if (!"".equals(centuryPartStr)) {
                    if (CENTURY_BEGINNING.equals(centuryPartStr)) {
                        startYear = (century - 1) * 100 + 1;
                        endYear = (century - 1) * 100 + 33;
                    } else if (CENTURY_MIDDLE.equals(centuryPartStr)) {
                        startYear = (century - 1) * 100 + 34;
                        endYear = (century - 1) * 100 + 66;
                    } else if (CENTURY_END.equals(centuryPartStr)) {
                        startYear = (century - 1) * 100 + 67;
                        endYear = century * 100;
                    } else if (CENTURY_FIRST_HALF.equals(centuryPartStr)) {
                        startYear = (century - 1) * 100 + 1;
                        endYear = (century - 1) * 100 + 50;
                    } else if (CENTURY_SECOND_HALF.equals(centuryPartStr)) {
                        startYear = (century - 1) * 100 + 51;
                        endYear = century * 100;
                    }
                }

                modifiedCentury = startStr + fieldStr + "[" + formatStartYear(startYear) + " TO " + formatEndYear(endYear) + "]";
            } else {
                modifiedCentury = startStr;
            }

            centuryRangeMatcher.appendReplacement(sb, modifiedCentury);
            res.setQueryModified(true);
        }
        centuryRangeMatcher.appendTail(sb);

        res.setFilteredQuery(sb.toString());
        return res;
    }
}
