package hu.kozadat.search.persistence.core;

import hu.kozadat.search.model.InstitutionNameAutoCompleteData;
import hu.scriptum.einfoszab.datamodel.Institution;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Repository;

import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

@Repository
public class InstitutionNameAutoCompleteOperations {

    private static final Logger LOGGER = LoggerFactory.getLogger(InstitutionNameAutoCompleteOperations.class);

    private static final String QUERY_DATAPROVIDERS_AND_EXTERNAL_INSTITUTES =
            "select "
                    + "dp.id" + " as id, "
                    + "dp.name" + " as label, "
                    + "dp.state" + " as status "
                    + "from" + " "
                    + "data_providers" + " dp "
                    + "where" + " "
                    + "dp.state" + " in " + "(" + "'A'" + ", " + "'R'" + ", " + "'I'" + ") and (lower(dp.name) like ? or lower(dp.short_name) like ?) "
                    + "union "
                    + "select "
                    + "ei.id" + " as id, "
                    + "ei.name" + " as label, "
                    + "null" + " as status "
                    + "from" + " "
                    + "external_institute" + " ei "
                    + "where" + " "
                    + "lower(ei.name) like ? "
                    + "order by label" + " "
                    + "limit ?";

    private static final String QUERY_DATAPROVIDERS_AND_EXTERNAL_INSTITUTES_ORDERED_BY_STATE =
            "(select "
                    + "dp.id as id, dp.name as label, dp.state as status, (case dp.state when 'A' then 0 when 'R' then 1 when 'I' then 3 else 99 end) as status_order "
                    + "from data_providers dp "
                    + "where dp.state in ('A', 'R', 'I') "
                    + "and (lower(dp.name) like lower(?) or lower(dp.short_name) like lower(?)) "
                    + "union "
                    + "select "
                    + "ei.id as id, ei.name as label, null as status, 2 as status_order "
                    + "from external_institute ei "
                    + "where lower(ei.name) like lower(?)) "
                    + "order by status_order asc, label asc "
                    + "limit ?";

    private static final String QUERY_DATAPROVIDERS_AND_EXTERNAL_INSTITUTES_ORDERED_BY_STATE_FORMAT =
            "(select "
                    + "dp.id as id, dp.name as label, dp.state as status, (case dp.state when 'A' then 0 when 'R' then 1 when 'I' then 3 else 99 end) as status_order "
                    + "from data_providers dp "
                    + "where dp.state in ('A', 'R', 'I') "
                    + "and (lower(dp.name) like lower('%s') or lower(dp.short_name) like lower('%s')) "
                    + "union "
                    + "select "
                    + "ei.id as id, ei.name as label, null as status, 2 as status_order "
                    + "from external_institute ei "
                    + "where lower(ei.name) like lower('%s')) "
                    + "order by status_order asc, label asc "
                    + "limit %d";

    private static final String COUNT_DATAPROVIDERS_AND_EXTERNAL_INSTITUTES =
            "select count(id) from ("
                    + "select "
                    + "dp.id" + " as id, "
                    + "dp.name" + " as label, "
                    + "dp.state" + " as status "
                    + "from" + " "
                    + "data_providers" + " dp "
                    + "where" + " "
                    + "dp.state" + " in " + "(" + "'A'" + ", " + "'R'" + ", " + "'I'" + ") and (lower(dp.name) like ? or lower(dp.short_name) like ?) "
                    + "union "
                    + "select "
                    + "ei.id" + " as id, "
                    + "ei.name" + " as label, "
                    + "null" + " as status "
                    + "from" + " "
                    + "external_institute" + " ei "
                    + "where" + " "
                    + "lower(ei.name) like ? "
                    + ")";

    private JdbcTemplate jdbcTemplate;

    protected String getAsValuePrefix(String valuePart) {
        if (valuePart != null) {
            return "%" + valuePart.toLowerCase() + "%";
        } else {
            return "%";
        }
    }

    public Set<InstitutionNameAutoCompleteData> findByPrefix(final String prefix, final int limit) {

        String sql;

        Set<InstitutionNameAutoCompleteData> result = new LinkedHashSet<>();

        String namePattern = getAsValuePrefix(trim(prefix, Institution.MAXLEN_NAME - 2));
        String shortNamePattern = getAsValuePrefix(trim(prefix, Institution.MAXLEN_SHORTNAME - 2));
        LOGGER.debug("findByPrefix(): namePattern={}, shortNamePattern={}", namePattern, shortNamePattern);

        // tnagy mantis issue fix 0001448
        sql = String.format(QUERY_DATAPROVIDERS_AND_EXTERNAL_INSTITUTES_ORDERED_BY_STATE_FORMAT, namePattern, shortNamePattern, namePattern, limit);

        List<InstitutionNameAutoCompleteData> institutes = jdbcTemplate.query(sql, new InstitutionNameAutoCompleteDataMapper());

        result.addAll(institutes);

        return result;
    }

    public int countByPrefix(String prefix) {

        String namePattern = getAsValuePrefix(trim(prefix, Institution.MAXLEN_NAME - 2));
        String shortNamePattern = getAsValuePrefix(trim(prefix, Institution.MAXLEN_SHORTNAME - 2));
        LOGGER.debug("countByPrefix(): namePattern={}, shortNamePattern={}", namePattern, shortNamePattern);

        int count = jdbcTemplate.queryForInt(COUNT_DATAPROVIDERS_AND_EXTERNAL_INSTITUTES, namePattern, shortNamePattern, namePattern);

        return count;
    }

    @Autowired
    public void setJdbcTemplate(JdbcTemplate jdbcTemplate) {
        this.jdbcTemplate = jdbcTemplate;
    }

    private String trim(String value, int maxLength) {
        if (value == null || value.length() == 0) {
            return value;
        }
        return value.substring(0, Math.min(value.length(), maxLength));
    }

}
