/**
 * Created by IntelliJ IDEA.
 * User: tnagy
 * Date: 2010.02.26.
 * Time: 14:37:31
 */
package hu.kozadat.search.persistence.recommendedsearch;

import hu.kozadat.search.exception.DictionarySearcherException;
import hu.kozadat.search.exception.TermParseException;
import hu.kozadat.search.persistence.DictionaryOperations;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.RAMDirectory;
import org.apache.lucene.util.Version;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.locks.ReentrantReadWriteLock;

/**
 * Class to search in dictionary specified by a {@link WordType}.
 * <p/>
 * There are several states the class instance can be:
 * <ul>
 * <li>1 - created </li>
 * <li>2 - initialized</li>
 * <li>3 - busy</li>
 * <li>4 - ready for search</li>
 * <li>5 - closed</li>
 * </ul>
 * <p/>
 * After calling the class constructor, the object states as created if a valid {@link WordType}
 * and a valid {@link hu.kozadat.search.persistence.recommendedsearch.DictionarySource} was given.
 * <p/>
 * Usage:
 * <ul>
 * <li>init (1)</li>
 * <li>build index (1)</li>
 * <li>search (*)</li>
 * <li>close (1)</li>
 * </ul>
 */
public class DictionarySearcher {

    private static final Logger LOGGER = LoggerFactory.getLogger(DictionarySearcher.class);

    private ReentrantReadWriteLock lock = new ReentrantReadWriteLock();

    public static final int STATE_CREATED = 1;
    public static final int STATE_WORDS_READ = 2;
    public static final int STATE_BUSY = 3;
    public static final int STATE_READY_FOR_SEARCH = 4;
    public static final int STATE_CLOSED = 5;

    private volatile int state = STATE_CREATED;

    private WordType wordType;
    private DictionaryOperations dictionaryOperations;
    private Map<String, String> words;

    private Analyzer analyzer;
    private Directory directory;
    private IndexWriter indexWriter;
    private IndexReader indexReader;
    private IndexSearcher indexSearcher;
    private float similarity = 0.5f;

    public void setWordType(WordType wordType) {
        this.wordType = wordType;
    }

    @Autowired
    public void setDictionaryOperations(DictionaryOperations dictionaryOperations) {
        this.dictionaryOperations = dictionaryOperations;
    }

    public int getState() {
        lock.readLock().lock();
        try {
            return state;
        } finally {
            lock.readLock().unlock();
        }
    }

    public void setSimilarity(float similarity) {
        this.similarity = similarity;
    }

    public void init() throws DictionarySearcherException {
        LOGGER.info("Init CALLED...");
        readDictionary();
        buildIndex();
    }

    public void readDictionary() throws DictionarySearcherException {
        long started = System.currentTimeMillis();
        LOGGER.info("Reading dictionary from " + dictionaryOperations);
        lock.writeLock().lock();
        try {
            if (state == STATE_CREATED) {
                if (dictionaryOperations != null) {
                    state = STATE_BUSY;
                    words = dictionaryOperations.readWords(wordType.getId());
                    analyzer = new StandardAnalyzer(Version.LUCENE_30);
                    directory = new RAMDirectory();
                    state = STATE_WORDS_READ;
                } else {
                    throw new DictionarySearcherException("Property 'dictionaryOperations' not set, cannot initialize.");
                }
            } else {
                throw new DictionarySearcherException("Cannot init, state is " + state);
            }
        } finally {
            lock.writeLock().unlock();
        }
        LOGGER.info("Reading words finished in " + (System.currentTimeMillis() - started) + " ms");
    }

    public void buildIndex() throws DictionarySearcherException {
        LOGGER.info("Building index...");
        lock.writeLock().lock();
        try {
            if (state == STATE_WORDS_READ) {
                if (words != null) {
                    Thread t = new Thread(new Runnable() {
                        public void run() {
                            long started = System.currentTimeMillis();
                            state = STATE_BUSY;
                            try {
                                indexWriter = new IndexWriter(directory, analyzer, /* create */ true, new IndexWriter.MaxFieldLength(DictionaryConstants.MAX_FIELD_LENGTH));

                                for (String word : words.keySet()) {
                                    Document document = new Document();
                                    document.add(new Field(DictionaryConstants.FIELD_NAME, word, Field.Store.YES, Field.Index.ANALYZED));
                                    indexWriter.addDocument(document);
                                }

                                indexWriter.close();

                                indexSearcher = new IndexSearcher(directory, /* read-only */ true);

                                indexReader = IndexReader.open(directory);

                                state = STATE_READY_FOR_SEARCH;
                                LOGGER.info("Index built in " + (System.currentTimeMillis() - started) + " ms");
                            } catch (IOException e) {
                                LOGGER.error(e.getMessage(), e);
                                state = STATE_WORDS_READ;
                                indexWriter = null;
                            }
                        }
                    });

                    t.start();
                } else {
                    throw new DictionarySearcherException("Cannot build index, not initialized properly.");
                }
            } else {
                throw new DictionarySearcherException("Cannot build index, wait until initialized.");
            }
        } finally {
            lock.writeLock().unlock();
        }
        LOGGER.info("Index built.");
    }


    public List<String> search(String queryString) throws DictionarySearcherException {
        lock.readLock().lock();
        try {
            if (state == STATE_READY_FOR_SEARCH) {
                Query query = new QueryParser(Version.LUCENE_30, DictionaryConstants.FIELD_NAME, analyzer).parse(queryString);
                ScoreDoc[] hits = indexSearcher.search(query, null, DictionaryConstants.HIT_PER_PAGE).scoreDocs;
                List<String> result = new LinkedList<String>();
                for (ScoreDoc scoreDoc : hits) {
                    int documentId = scoreDoc.doc;
                    Document document = indexSearcher.doc(documentId);
                    result.add(document.get(DictionaryConstants.FIELD_NAME));
                }
                return result;
            } else {
                throw new DictionarySearcherException("DictionarySearch is busy, try later.");
            }
        } catch (IOException e) {
            throw new DictionarySearcherException(e.getMessage(), e);
        } catch (ParseException e) {
            throw new TermParseException(e.getMessage(), e);
        } finally {
            lock.readLock().unlock();
        }
    }

    public Map<String, List<String>> fuzzySearch(String queryString) throws DictionarySearcherException {
        lock.readLock().lock();
        try {
            if (state == STATE_READY_FOR_SEARCH) {
                Query parsedQuery = new QueryParser(Version.LUCENE_30, DictionaryConstants.FIELD_NAME, analyzer).parse(queryString);
                Query rewrittenQuery = parsedQuery.rewrite(indexReader);
                Set<Term> terms = new HashSet<Term>();
                rewrittenQuery.extractTerms(terms);
                Map<String, List<String>> result = new TreeMap<String, List<String>>();
                for (Term term : terms) {
                    Query query = new FuzzyQueryParser(Version.LUCENE_30, DictionaryConstants.FIELD_NAME, analyzer).getFuzzyQuery(term.text(), similarity);
                    ScoreDoc[] hits = indexSearcher.search(query, null, DictionaryConstants.HIT_PER_PAGE).scoreDocs;
                    List<String> resultList = new LinkedList<String>();
                    for (ScoreDoc scoreDoc : hits) {
                        int documentId = scoreDoc.doc;
                        Document document = indexSearcher.doc(documentId);
                        String word = document.get(DictionaryConstants.FIELD_NAME);
                        if (words.get(word) != null) {
                            resultList.add(words.get(word));
                        } else {
                            resultList.add(word);
                        }
                    }
                    result.put(term.text(), resultList);
                }
                return result;
            } else {
                throw new DictionarySearcherException("DictionarySearch is busy, try later.");
            }
        } catch (IOException e) {
            throw new DictionarySearcherException(e.getMessage(), e);
        } catch (ParseException e) {
            throw new TermParseException(e.getMessage(), e);
        } finally {
            lock.readLock().unlock();
        }
    }

    public void close() throws DictionarySearcherException {
        lock.writeLock().lock();
        try {
            indexSearcher.close();
            indexReader.close();
            directory.close();
            state = STATE_CLOSED;
        } catch (IOException e) {
            throw new DictionarySearcherException(e.getMessage(), e);
        } finally {
            lock.writeLock().unlock();
        }
    }
}
