package hu.kozadat.search.presentation.simplesearch;

import hu.kozadat.search.exception.CSVExportException;
import hu.kozadat.search.model.IndexerSearchResult;
import hu.kozadat.search.model.IndexerSearchResultItem;
import hu.kozadat.search.persistence.DataProviderOperations;
import hu.kozadat.search.presentation.CacheAwareController;
import hu.kozadat.search.presentation.UrlUtils;
import hu.kozadat.search.service.SearchUIConfiguration;
import hu.kozadat.search.service.export.CSVExportResult;
import hu.kozadat.search.service.export.ExportService;
import hu.kozadat.search.service.log.SearchLogger;
import hu.kozadat.search.service.recommendedsearch.RecommendedSearchService;
import hu.kozadat.search.service.searchstatistic.SearchStatisticService;
import hu.kozadat.search.service.simplesearch.SimpleSearch;
import hu.scriptum.einfoszab.datamodel.DataProvider;
import hu.scriptum.einfoszab.datamodel.Schema;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.support.PagedListHolder;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Controller;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.servlet.ModelAndView;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;
import javax.validation.Valid;
import java.text.SimpleDateFormat;
import java.util.*;

/**
 * Egyszeru kereses form kontroller-e.
 *
 * @author Csaba Horvath, Scriptum Inc.
 * @version $Id$
 */
@Controller
public class SimpleSearchController implements CacheAwareController, InitializingBean {

    private static final Logger LOGGER = LoggerFactory.getLogger(SimpleSearchController.class);

    private static final String SESSION_ATTRIBUTE_INDEXER_SEARCH_RESULT = "indexerSearchResult";
    private static final String SESSION_ATTRIBUTE_INSTITUTION_RECORD_INDEXER_SEARCH_RESULT = "institutionRecordIndexerSearchResult";

    private static final String SESSION_ATTRIBUTE_PAGED_SEARCH_RESULT = "pagedSearchResult";
    private static final String SESSION_ATTRIBUTE_PAGED_INSTITUTION_RECORD_SEARCH_RESULT = "pagedInstitutionRecordSearchResult";

    private static final String HTTP_HEADER_NAME_CONTENT_DISPOSITION = "Content-Disposition";
    private static final String HTTP_HEADER_VALUE_ATTACHMENT_FILENAME_PART = "attachment; filename=";
    private static final SimpleDateFormat EXPORT_DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss");

    private SearchLogger searchLogger;
    private SimpleSearch simpleSearch;
    private RecommendedSearchService recommendedSearchService;
    private SearchStatisticService searchStatisticService;
    private DataProviderOperations dataProviderOperations;
    private SearchUIConfiguration searchUIConfiguration;
    private ExportService exportService;

    private List<Schema> schemas;
    private Map<String, Long> schemaIdMap = new TreeMap<String, Long>();
    private boolean checkCritical = false;

    public void afterPropertiesSet() throws Exception {
        this.checkCritical = searchUIConfiguration.getSearchCheckCriticalWords();

        this.schemas = (List<Schema>) dataProviderOperations.getSupportedSchemas();
        for (Schema schema : schemas) {
            if (schema.getPrefix() != null && !"".equals(schema.getPrefix())) {
                schemaIdMap.put(schema.getPrefix(), schema.getId());
            }
        }
    }

    protected void clearSessionAttributes(HttpSession session) {
        if (session != null) {
            session.removeAttribute(SESSION_ATTRIBUTE_INDEXER_SEARCH_RESULT);
            session.removeAttribute(SESSION_ATTRIBUTE_PAGED_SEARCH_RESULT);
            session.removeAttribute(SESSION_ATTRIBUTE_INSTITUTION_RECORD_INDEXER_SEARCH_RESULT);
            session.removeAttribute(SESSION_ATTRIBUTE_PAGED_INSTITUTION_RECORD_SEARCH_RESULT);
        }
    }

    @ModelAttribute("simpleSearchFormBackingBean")
    protected SimpleSearchFormBackingBean createSimpleSearchFormBackingBean() {
        SimpleSearchFormBackingBean form = new SimpleSearchFormBackingBean();
        List<String> selectedSchemas = new LinkedList<String>();
        for (Schema schema : schemas) {
            selectedSchemas.add(Long.toString(schema.getId()));
        }
        form.setSema(selectedSchemas);
        form.setAvailableSchemas(schemas);
        return form;
    }

    @InitBinder
    protected void initBinder(WebDataBinder binder) {
        SimpleSearchValidator validator = new SimpleSearchValidator();
        validator.setSearchTextMinLength(searchUIConfiguration.getSearchTextMinLength());
        validator.setSearchTextMaxLength(searchUIConfiguration.getSearchTextMaxLength());
        binder.setValidator(validator);
    }

    @RequestMapping(value = "/", method = RequestMethod.GET)
    public ModelAndView showForm(HttpServletRequest httpServletRequest) throws Exception {
        HttpSession session = httpServletRequest.getSession();

        clearSessionAttributes(session);

        String viewName = "simpleSearch";
        ModelAndView modelAndView = new ModelAndView(viewName);
        return modelAndView;
    }

    @RequestMapping(value = "/talalatok")
    public ModelAndView searchAndShowResult(HttpServletRequest httpServletRequest, @RequestParam(value = "oldal", required = false) Integer page, @RequestParam(value = CacheAwareController.PARAM_NAME_CLEAR_CACHE, required = false) String clearCache, @ModelAttribute("simpleSearchFormBackingBean") @Valid SimpleSearchFormBackingBean form, BindingResult bindingResult) throws Exception {
        HttpSession session = httpServletRequest.getSession();
        String sessionId = httpServletRequest.getSession().getId();

        if (CacheAwareController.PARAM_VALUE_CLEAR_CACHE.equals(clearCache)) {
            clearSessionAttributes(session);
            String viewName = UrlUtils.redirectTo(UrlUtils.excludeParam("/talalatok", httpServletRequest, CacheAwareController.PARAM_NAME_CLEAR_CACHE));
            return new ModelAndView(viewName);
        }

        if (bindingResult.hasErrors()) {

            clearSessionAttributes(session);

            String viewName = "simpleSearchResult";
            ModelAndView modelAndView = new ModelAndView(viewName);
            modelAndView.addObject("simpleSearchFormBackingBean", form);
            return modelAndView;
        }

        form.setDataProviderName(null);

        IndexerSearchResult indexerSearchResult;
        PagedListHolder<IndexerSearchResultItem> pagedIndexerSearchResult;

        indexerSearchResult = (IndexerSearchResult) session.getAttribute(SESSION_ATTRIBUTE_INDEXER_SEARCH_RESULT);
        pagedIndexerSearchResult = (PagedListHolder<IndexerSearchResultItem>) session.getAttribute(SESSION_ATTRIBUTE_PAGED_SEARCH_RESULT);

        boolean formChange = false;
        if (indexerSearchResult != null) {
            formChange = indexerSearchResult.hasFormChange(form);
        }

        if (indexerSearchResult == null || pagedIndexerSearchResult == null || formChange) {

            boolean containsCritical = checkCritical && recommendedSearchService.containsCriticalWord(form.getSzoveg());

            List<String> selectedSchemas = new LinkedList<String>();
            if (form.getSema() == null || form.getSema().isEmpty()) {
                form.setSema(selectedSchemas);
                for (Schema schema : form.getAvailableSchemas()) {
                    selectedSchemas.add(Long.toString(schema.getId()));
                }
            } else if (form.getSema().size() > 0) {
                if (form.getSema().size() != form.getAvailableSchemas().size()) {
                    selectedSchemas.addAll(form.getSema());
                }
            }

            indexerSearchResult = simpleSearch.search(form.getSzoveg(), selectedSchemas, sessionId);
            indexerSearchResult.setFormHashCode(form);

            String searchExpression = indexerSearchResult.getQuery();
            int resultSize = indexerSearchResult.getResultSize();

            if (resultSize == 0) {
                if (!containsCritical) {
                    searchLogger.addSimpleSearchLog(httpServletRequest.getRemoteHost(), form.getSzoveg(), null, 0, 0);
                }
                ModelAndView noResultView = new ModelAndView("noSearchResult");
                noResultView.addObject("szoveg", form.getSzoveg());
                return noResultView;
            } else if (!containsCritical) {
                searchStatisticService.storeQuery(searchExpression);
            }

            simpleSearch.fillResult(indexerSearchResult, 0, resultSize);
            simpleSearch.close(sessionId);

            if (!containsCritical) {

                searchLogger.addSimpleSearchLog(httpServletRequest.getRemoteHost(),
                        form.getSzoveg(),
                        null,
                        indexerSearchResult.getExecutionTime().intValue() + indexerSearchResult.getProcessTime().intValue(),
                        indexerSearchResult.getResultSize());
            }

            pagedIndexerSearchResult = new PagedListHolder<IndexerSearchResultItem>();
            pagedIndexerSearchResult.setPageSize(searchUIConfiguration.getSearchResultPageSize());
            pagedIndexerSearchResult.setSource(indexerSearchResult.getFirstItems());

            httpServletRequest.getSession().setAttribute(SESSION_ATTRIBUTE_INDEXER_SEARCH_RESULT, indexerSearchResult);
            httpServletRequest.getSession().setAttribute(SESSION_ATTRIBUTE_PAGED_SEARCH_RESULT, pagedIndexerSearchResult);
        }

        if (page == null || page < 1) {
            page = 1;
        }
        if (page > pagedIndexerSearchResult.getPageCount()) {
            page = pagedIndexerSearchResult.getPageCount();
        }
        pagedIndexerSearchResult.setPage(page - 1);

        String viewName = "simpleSearchResult";
        ModelAndView modelAndView = new ModelAndView(viewName);

        modelAndView.addObject("searchResult", indexerSearchResult);
        modelAndView.addObject("pagedSearchResult", pagedIndexerSearchResult);

        return modelAndView;

    }

    @RequestMapping("/talalatok/{dataProviderId}")
    public ModelAndView searchAndShowResultByDataProvider(HttpServletRequest httpServletRequest, @PathVariable("dataProviderId") String dataProviderIdStr, @RequestParam(value = "oldal", required = false) Integer page, @RequestParam(value = CacheAwareController.PARAM_NAME_CLEAR_CACHE, required = false) String clearCache, @ModelAttribute("simpleSearchFormBackingBean") @Valid SimpleSearchFormBackingBean form, BindingResult bindingResult) throws Exception {

        HttpSession session = httpServletRequest.getSession();
        String sessionId = session.getId();

        if (CacheAwareController.PARAM_VALUE_CLEAR_CACHE.equals(clearCache)) {
            clearSessionAttributes(session);
            String viewName = UrlUtils.redirectTo(UrlUtils.excludeParam("/talalatok/" + dataProviderIdStr, httpServletRequest, CacheAwareController.PARAM_NAME_CLEAR_CACHE));
            return new ModelAndView(viewName);
        }

        boolean containsCritical = false;
        if (form.getSzoveg() != null && !"".equals(form.getSzoveg())) {
            containsCritical = checkCritical && recommendedSearchService.containsCriticalWord(form.getSzoveg());
        }

        Long dataProviderId = null;
        try {
            dataProviderId = Long.parseLong(dataProviderIdStr);
        } catch (NumberFormatException e) {
            LOGGER.warn("InstitutionRecordsController.showResultByDataProvider(): cannot parse '" + dataProviderIdStr + "', " + e.getMessage());
        }

        if (dataProviderId == null) {
            return new ModelAndView("redirect:/");
        }

        form.setDataProviderId(dataProviderId);

        IndexerSearchResult institutionRecordIndexerSearchResult = (IndexerSearchResult) session.getAttribute(SESSION_ATTRIBUTE_INSTITUTION_RECORD_INDEXER_SEARCH_RESULT);
        PagedListHolder<IndexerSearchResultItem> pagedInstitutionRecordSearchResult = (PagedListHolder<IndexerSearchResultItem>) session.getAttribute(SESSION_ATTRIBUTE_PAGED_INSTITUTION_RECORD_SEARCH_RESULT);
        IndexerSearchResult indexerSearchResult = (IndexerSearchResult) session.getAttribute(SESSION_ATTRIBUTE_INDEXER_SEARCH_RESULT);

        boolean doSearch;
        if (institutionRecordIndexerSearchResult == null || pagedInstitutionRecordSearchResult == null) {
            if (indexerSearchResult != null && indexerSearchResult.getItemsByDataProviderIds().containsKey(dataProviderId)) {
                institutionRecordIndexerSearchResult = indexerSearchResult.getItemsByDataProviderAsSearchResult(dataProviderId);
                institutionRecordIndexerSearchResult.setFormHashCode(indexerSearchResult.getFormHashCode());

                session.setAttribute(SESSION_ATTRIBUTE_INSTITUTION_RECORD_INDEXER_SEARCH_RESULT, institutionRecordIndexerSearchResult);

                pagedInstitutionRecordSearchResult = new PagedListHolder<IndexerSearchResultItem>();
                pagedInstitutionRecordSearchResult.setPageSize(searchUIConfiguration.getSearchResultPageSize());
                if (institutionRecordIndexerSearchResult.getItemsByDataProviderIds().containsKey(dataProviderId)) {
                    Collection<IndexerSearchResultItem> collection = institutionRecordIndexerSearchResult.getItemsByDataProviderIds().get(dataProviderId);
                    pagedInstitutionRecordSearchResult.setSource(new LinkedList<>(collection));
                }

                httpServletRequest.getSession().setAttribute(SESSION_ATTRIBUTE_PAGED_INSTITUTION_RECORD_SEARCH_RESULT, pagedInstitutionRecordSearchResult);
            }
        }
        doSearch = institutionRecordIndexerSearchResult == null || institutionRecordIndexerSearchResult.hasFormChange(form);

        // tnagy search if first request or conditions changed
        if (doSearch) {

            List<String> selectedSchemas = new LinkedList<String>();
            if (form.getSema() == null || form.getSema().isEmpty()) {
                form.setSema(selectedSchemas);
                for (Schema schema : form.getAvailableSchemas()) {
                    selectedSchemas.add(Long.toString(schema.getId()));
                }
            } else if (form.getSema().size() > 0) {
                if (form.getSema().size() != form.getAvailableSchemas().size()) {
                    selectedSchemas.addAll(form.getSema());
                }
            }

            institutionRecordIndexerSearchResult = simpleSearch.searchInstitution(dataProviderId, form.getSzoveg(), selectedSchemas, sessionId);
            institutionRecordIndexerSearchResult.setFormHashCode(form);
            simpleSearch.close(sessionId);

            session.setAttribute(SESSION_ATTRIBUTE_INSTITUTION_RECORD_INDEXER_SEARCH_RESULT, institutionRecordIndexerSearchResult);

            if (institutionRecordIndexerSearchResult.getResultSize() > 0) {

                IndexerSearchResultItem firstItem = institutionRecordIndexerSearchResult.getItems().get(0);

                if (!containsCritical) {
                    searchLogger.addInstitutionSearchLog(
                            httpServletRequest.getRemoteAddr(),
                            firstItem.getDataProviderName(),
                            firstItem.getDataProviderId(),
                            institutionRecordIndexerSearchResult.getProcessTime().intValue(),
                            institutionRecordIndexerSearchResult.getResultSize());
                }

            }

            pagedInstitutionRecordSearchResult = new PagedListHolder<IndexerSearchResultItem>();
            pagedInstitutionRecordSearchResult.setPageSize(searchUIConfiguration.getSearchResultPageSize());
            if (institutionRecordIndexerSearchResult.getItemsByDataProviderIds().containsKey(dataProviderId)) {
                Collection<IndexerSearchResultItem> collection = institutionRecordIndexerSearchResult.getItemsByDataProviderIds().get(dataProviderId);
                pagedInstitutionRecordSearchResult.setSource(new LinkedList<>(collection));
            }

            httpServletRequest.getSession().setAttribute(SESSION_ATTRIBUTE_PAGED_INSTITUTION_RECORD_SEARCH_RESULT, pagedInstitutionRecordSearchResult);

        }

        // tnagy get dataprovider name from result or from dataprovider db table if result was empty
        IndexerSearchResultItem firstItem = null;
        if (institutionRecordIndexerSearchResult != null && institutionRecordIndexerSearchResult.getItemsByDataProviderIds().get(dataProviderId) != null) {
            Collection<IndexerSearchResultItem> collection = institutionRecordIndexerSearchResult.getItemsByDataProviderIds().get(dataProviderId);
            firstItem = new LinkedList<>(collection).get(0);
            if (firstItem != null) {
                form.setDataProviderName(firstItem.getDataProviderName());
            }
        }
        if (firstItem == null) {
            DataProvider dataProvider = dataProviderOperations.getDataProvider(dataProviderId);
            if (dataProvider == null) {
                return new ModelAndView("redirect:/");
            }
            form.setDataProviderName(dataProvider.getInstitution().getName());
        }

        // tnagy set page
        if (page == null || page < 1) {
            page = 1;
        }
        if (page > pagedInstitutionRecordSearchResult.getPageCount()) {
            page = pagedInstitutionRecordSearchResult.getPageCount();
        }
        pagedInstitutionRecordSearchResult.setPage(page - 1);


        String viewName = "institutionRecordSearchResult";
        ModelAndView modelAndView = new ModelAndView(viewName);

        modelAndView.addObject(SESSION_ATTRIBUTE_INDEXER_SEARCH_RESULT, institutionRecordIndexerSearchResult);
        modelAndView.addObject(SESSION_ATTRIBUTE_PAGED_INSTITUTION_RECORD_SEARCH_RESULT, pagedInstitutionRecordSearchResult);
        modelAndView.addObject("dataProviderId", dataProviderId);
        return modelAndView;

    }

    @RequestMapping(value = "/export/csv", method = RequestMethod.GET)
    public HttpEntity<byte[]> exportRecords(HttpSession session) throws CSVExportException {
        IndexerSearchResult indexerSearchResult = (IndexerSearchResult) session.getAttribute(SESSION_ATTRIBUTE_INDEXER_SEARCH_RESULT);
        if (indexerSearchResult == null) {
            LOGGER.warn("exportRecords(): IndexerSearchResult not found in session.");
        }
        return exportResult(indexerSearchResult);
    }

    @RequestMapping(value = "/export/institution/csv", method = RequestMethod.GET)
    public HttpEntity<byte[]> exportInstitutionRecords(HttpSession session) throws CSVExportException {
        IndexerSearchResult indexerSearchResult = (IndexerSearchResult) session.getAttribute(SESSION_ATTRIBUTE_INSTITUTION_RECORD_INDEXER_SEARCH_RESULT);
        if (indexerSearchResult == null) {
            LOGGER.warn("exportInstitutionRecords(): IndexerSearchResult not found in session.");
        }
        return exportResult(indexerSearchResult);
    }

    private HttpEntity<byte[]> exportResult(IndexerSearchResult indexerSearchResult) throws CSVExportException {
        HttpHeaders headers = new HttpHeaders();

        long start = System.currentTimeMillis();

        CSVExportResult csvExportResult = exportService.exportToCSV(indexerSearchResult);

        String fileName = "KozadatKereso_" + EXPORT_DATE_FORMAT.format(new Date()) + ".csv";

        headers.setContentLength(csvExportResult.getLength());
        headers.setContentType(new MediaType("text", "plain", csvExportResult.getCharset()));
        headers.set(HTTP_HEADER_NAME_CONTENT_DISPOSITION, HTTP_HEADER_VALUE_ATTACHMENT_FILENAME_PART + "\"" + fileName.replace(" ", "_") + "\"");

        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("exportResult(): Export finished in {} ms.", (System.currentTimeMillis() - start));
        }

        return new HttpEntity<>(csvExportResult.getContent(), headers);
    }

    @Autowired
    public void setSearchLogger(SearchLogger searchLogger) {
        this.searchLogger = searchLogger;
    }

    @Autowired
    public void setSimpleSearch(SimpleSearch simpleSearch) {
        this.simpleSearch = simpleSearch;
    }

    @Autowired
    public void setRecommendedSearchService(RecommendedSearchService recommendedSearchService) {
        this.recommendedSearchService = recommendedSearchService;
    }

    @Autowired
    public void setSearchStatisticService(SearchStatisticService searchStatisticService) {
        this.searchStatisticService = searchStatisticService;
    }

    @Autowired
    public void setDataProviderOperations(DataProviderOperations dataProviderOperations) {
        this.dataProviderOperations = dataProviderOperations;
    }

    @Autowired
    public void setSearchUIConfiguration(SearchUIConfiguration searchUIConfiguration) {
        this.searchUIConfiguration = searchUIConfiguration;
    }

    @Autowired
    public void setExportService(ExportService exportService) {
        this.exportService = exportService;
    }
}
