package hu.kozadat.search.service.interactivesearch;

import hu.kozadat.search.base.preprocessor.QueryPreprocessorManager;
import hu.kozadat.search.base.preprocessor.QueryPreprocessorResult;
import hu.kozadat.search.interactive.InteractiveSearchExpression;
import hu.kozadat.search.model.IndexerSearchResult;
import hu.kozadat.search.model.IndexerSearchResultItem;
import hu.kozadat.search.service.IndexerSearchResultParser;
import hu.kozadat.search.service.SearchUIConfiguration;
import hu.scriptum.nda.api.MDcSearch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

@Service("interactiveSearch")
public class InteractiveSearch implements InitializingBean {

    private static final Logger LOGGER = LoggerFactory.getLogger(InteractiveSearch.class);

    private SearchUIConfiguration searchUIConfiguration;
    private MDcSearch searchService;
    private int maxResultCount = 10000;
    private QueryPreprocessorManager queryPreprocessor;

    public IndexerSearchResult search(InteractiveSearchExpression expression, String sessionId) {

        IndexerSearchResult result = new IndexerSearchResult();

        QueryPreprocessorResult preprocessorResult = queryPreprocessor.preprocess(expression.toQueryString() != null ? expression.toQueryString().toLowerCase() : expression.toQueryString(), true);
        String searchExpression = preprocessorResult.getFilteredQuery();

        try {

            long time = System.currentTimeMillis();

            int resultSize = 0;
            try {
                resultSize = searchService.executeQuery(sessionId, searchExpression);
            } catch (Exception e) {
                LOGGER.error("Simple search failed: " + e.getMessage(), e);
            }

            result.setResultSize(resultSize);

            long queryElapsedTime = System.currentTimeMillis() - time;
            result.setExecutionTime(queryElapsedTime);

            time = System.currentTimeMillis();

            if (resultSize > 0) {

                String[] hits = searchService.getHits(sessionId, 0, resultSize < maxResultCount ? resultSize : maxResultCount);

                IndexerSearchResultParser indexerSearchResultParser = new IndexerSearchResultParser();
                for (String hit : hits) {

                    IndexerSearchResultItem indexerSearchResultItem = indexerSearchResultParser.parse(hit);
                    result.addItem(indexerSearchResultItem);

                }

            }
            long resultProcessTime = System.currentTimeMillis() - time;
            result.setProcessTime(resultProcessTime);

        } catch (Exception ex) {
            result.setResultSize(0);
            result.setProcessTime(0L);
        }

        result.setCharactersRemovedFromQuery(preprocessorResult.getRemovedCharsAsString());

        return result;
    }

    public void close(String sessionId) {
        searchService.closeSession(sessionId);
    }

    @Autowired
    public void setSearchService(MDcSearch searchService) {
        this.searchService = searchService;
    }

    @Autowired
    public void setQueryPreprocessor(QueryPreprocessorManager queryPreprocessor) {
        this.queryPreprocessor = queryPreprocessor;
    }

    @Autowired
    public void setSearchUIConfiguration(SearchUIConfiguration searchUIConfiguration) {
        this.searchUIConfiguration = searchUIConfiguration;
    }

    public void afterPropertiesSet() throws Exception {
        this.maxResultCount = searchUIConfiguration.getInteractiveSearchMaxResultCount();
    }
}
